/*
*  This script is the core of the Bookmark Duplicate Detector extension.
*
* Author: Stéphane BERTIN
* http://bookmarkdd.mozdev.org
*
*/


// Initiate bookmarks service to access all bookmarks methods
var bmsvc = Components.classes["@mozilla.org/browser/nav-bookmarks-service;1"]
    .getService(Components.interfaces.nsINavBookmarksService);
// Initiate livemarks service to access all livemarks methods
var livemarkService = Components.classes["@mozilla.org/browser/livemark-service;2"]
    .getService(Components.interfaces.nsILivemarkService);
var Application = Components.classes["@mozilla.org/fuel/application;1"]
    .getService(Components.interfaces.fuelIApplication);
var alertsService = Components.classes["@mozilla.org/alerts-service;1"]
    .getService(Components.interfaces.nsIAlertsService);

var Bdd_Alert_listener = {
    observe: function(subject, topic, cookie) {
	if (cookie) {
	    var Acookie = cookie.split("\+,\+");
	} else { return; }
	//FIXME : if user click and accept to keep too slowly (after popup is gone) bookmark is discarded.
	// By default, the bookmark is removed except if user clicks on Alert window and answers he wants to keep it.
	var removeIt = true; //Application.storage.get("Bdd_delete_bookmark_"+Acookie[0], true);
	if (topic == "alertclickcallback") {
	    var result = BddConfirm(Acookie[2],true);
	    if (typeof(result) == "undefined") { result=true; }
	    if(result) {
		Application.storage.set("Bdd_delete_bookmark_"+Acookie[0], false);
		if(Bookmarkdd.DEBUG) { Bookmarkdd.myDump("Confirmation to keep duplicate Bookmark ("+Acookie[0]+")\nUri="+Acookie[1]); }
	    } else {
		if(Bookmarkdd.DEBUG) { Bookmarkdd.myDump("Confirmation to delete duplicate Bookmark ("+Acookie[0]+")\nUri="+Acookie[1]); }
	    }
	}
	if (topic == "alertfinished") {
	    removeIt = Application.storage.get("Bdd_delete_bookmark_"+Acookie[0], true);
	    if (removeIt) {
		bmsvc.removeItem(Acookie[0]);
		if(Bookmarkdd.DEBUG) { Bookmarkdd.myDump("Duplicate Bookmark deleted ("+Acookie[0]+")\nUri="+Acookie[1]); }
	    } else {
		if(Bookmarkdd.DEBUG) { Bookmarkdd.myDump("Duplicate Bookmark kept ("+Acookie[0]+")\nUri="+Acookie[1]); }
	    }
	}
	// ELSE topic != [alertclickcallback,alertfinished]
	return;
    }
}
    
    


// nsINavBookmarkObserver : Listener for Bookmarkdd extension
// Observer to listen changes related to bookmarks
var Bdd_bookmarkListener = {
    onBeginUpdateBatch: function() {
	// function is not called !
	//Bookmarkdd.myDump("onBeginUpdateBatch");
    },
    onEndUpdateBatch: function() {
	// function is not called !
	//Bookmarkdd.myDump("onEndUpdateBatch");
    },
    onItemAdded: function(aItemId, aFolder, aIndex) {
	Bookmarkdd.doItemAdded(aItemId,aFolder,aIndex);
    },
    onItemRemoved: function(aItemId, aFolder, aIndex) {},
    onItemChanged: function(aBookmarkId, aProperty, aIsAnnotationProperty, aValue) {
	//Bookmarkdd.doSomething(aBookmarkId,aValue,"Changed");
    },
    onItemVisited: function(aBookmarkId, aVisitID, time) {},
    onItemMoved: function(aItemId, aOldParent, aOldIndex, aNewParent, aNewIndex) {
	// If a bookmark is moved in the same folder do not try to detect duplicates
	// Move in same folder happened with 'sort in folder'
	if (aOldParent != aNewParent) { Bookmarkdd.doItemAdded(aItemId,aNewParent,aNewIndex); }
    },
    QueryInterface: function(iid) {
	if (iid.equals(Ci.nsINavBookmarkObserver) || iid.equals(Ci.nsISupports)) {
	    return this;
	}
	throw Cr.NS_ERROR_NO_INTERFACE;
    }
};

// Extension Object
var Bookmarkdd = {

    // Property used with BddConfirmWindow.xul to get user choice
    userResponse: false,

    // Property used to trace some events during development
    DEBUG: true,

    // Property used to specify if a sort is processing on Bookmarks
    temporaryDisabled: false,
 
    _stringLocalization: Components.classes["@mozilla.org/intl/stringbundle;1"]
    .getService(Components.interfaces.nsIStringBundleService)
    .createBundle("chrome://duplicates/locale/duplicates.properties"),

    // Method used to load Bdd Observer
    onLoad: function() {
	bmsvc.addObserver(Bdd_bookmarkListener, false);
    },
 
    // Method used to Unload Bdd Observer
    onUnLoad: function() {
	bmsvc.removeObserver(Bdd_bookmarkListener);
    },

    // Method called by Bdd Observer when an item is added in bookmarks
    doItemAdded: function(aItemId,aFolderId,aIndex) {

	// Do not detect duplicate when sortFolderByName is processing
	Bookmarkdd.temporaryDisabled = Application.storage.get("Bdd_detectiondisabled", false);
	// 
	if (Bookmarkdd.temporaryDisabled) {
	    if(this.DEBUG) { this.myDump("temporaryDisabled is true"); }
	    return false;
	}
	//if(this.DEBUG) { this.myDump("temporaryDisabled is false"); }

	var thisURI = bmsvc.getBookmarkURI(aItemId);
	var itemType = bmsvc.getItemType(aItemId);
	//TYPE_BOOKMARK 	 1 	This is an index for type "Bookmark".
	//TYPE_FOLDER 	2 	This is an index for type "Folder".
	//TYPE_SEPARATOR 	3 	This is an index for type "Separator".
	//TYPE_DYNAMIC_CONTAINER  4 	This is an index for type "Dynamic Container".

	if (itemType == 1) {
	    var livemarkCheck = livemarkService.isLivemark(aFolderId); //if livemark return true
	    if (!livemarkCheck && !Bookmarkdd.isItaTag(aItemId)) {
		var bookmarksIdsForURI = bmsvc.getBookmarkIdsForURI(thisURI, {});
	       	// length>1 then URL is duplicated
		if (bookmarksIdsForURI.length > 1 && UserPrefs.getPref("string","Miscellaneous.DuplicateDetectionType") != "DoNotDetect") {
		    Bookmarkdd.manageDuplicate(aItemId,thisURI,bookmarksIdsForURI);
		}
		// TODO Add new function to search on partial match URL : detect on same domain name

	    } // else the item is a Livemark
	} else { // The item is not a bookmark
	    if(this.DEBUG) { this.myDump("Bdd added item is :\nType="+itemType+"\nUri="+thisURI.spec); }
	}

    },
    /*
     * Method to manage duplicate bookmark
     */
    manageDuplicate : function(aItemId,thisURI,bookmarksIdsForURI) {
	//TODO : factoring of code to avoid Bdd_delete_bookmark_

	// Function used to filter aItemId and Tagged bookmarks in the array
	function deleteElt(element, index, array) {
	    if (element == aItemId) { return false; }
	    if (Bookmarkdd.isItaTag(element)) { return false; }
	    return true;
	}
	var truncatedArray=bookmarksIdsForURI.filter(deleteElt);

	// Test to see if all elements have not been deleted by deleteElt()
	if (truncatedArray.length < 1) { return false; }

	var parentFoldersId = truncatedArray.map(bmsvc.getFolderIdForItem);
	var parentTitle = parentFoldersId.map(this.findPath); 

	var message = Bookmarkdd._stringLocalization.GetStringFromName("alreadyIn-Folder")+"\n";
	message += parentTitle.join("\n")+"\n";
	message += Bookmarkdd._stringLocalization.GetStringFromName("alreadyIn-Confirm");

	//TODO Add possibility to disable Alert message for import bookmark function
	var User_pref_DetectionType = UserPrefs.getPref("string","Miscellaneous.DuplicateDetectionType");
	if(User_pref_DetectionType == "AskConfirmation") {
	    // result =
	    // true -> keep duplicate bookmark if problem keep duplicate
	    // false -> delete item because it is already created when the method is called by Observer
	    var result = BddConfirm(message,true);
	    if (typeof(result) == "undefined") { result=true; }
	    if(result) {
		if(this.DEBUG) { this.myDump("Confirmation to keep duplicate Bookmark\nUri="+thisURI.spec); }
	    } else {
		if(this.DEBUG) { this.myDump("Confirmation to delete duplicate Bookmark\nUri="+thisURI.spec); }
		bmsvc.removeItem(aItemId);
	    }
	} else if(User_pref_DetectionType == "DiscardWithPopupAlert") {
	    const TRUNCATENUM = UserPrefs.getPref("int","URL.truncateNum");
	    var TruncatedUrl = thisURI.spec.substr(0,TRUNCATENUM);
	    var Ainfo = new Array(aItemId, TruncatedUrl, message);
	    var cookie = Ainfo.join("\+,\+");
	    Application.storage.set("Bdd_delete_bookmark_"+aItemId, true);
	    alertsService.showAlertNotification("chrome://duplicates/skin/duplicates-icon.png",  Bookmarkdd._stringLocalization.GetStringFromName("Bookmarkdd.notifier.discard"), TruncatedUrl, true, cookie, Bdd_Alert_listener);
	} else if(User_pref_DetectionType == "SilentlyDiscard") {
	    if(this.DEBUG) { this.myDump("Silently discard duplicate Bookmark\nUri="+thisURI.spec); }
	    bmsvc.removeItem(aItemId);
	} else {
	    if(this.DEBUG) { this.myDump("Value not supported for option DuplicateDetectionType = "+User_pref_DetectionType); }
	    return false;
	}
	return true;

    },
    /*
     * Method used to get the path of a folder Id.
     * If argument is not a folder -> return the ItemTitle
     * input argument : 
     * afolderId = a Folder Id
     * ouput : Return a chain of URL Parent Folder Name
     */
    findPath : function(afolderId) {
	var folderPath = new Array();
	var boucle = true;
	var itemType = bmsvc.getItemType(afolderId);

	// If argument is not a folder we do not try to find folder parent but return the ItemTitle
	if (itemType != 2) {
	    boucle = false;
	    //alert("FindPath: argument is not a Folder = "+bmsvc.getItemTitle(afolderId));
	} else {
	    folderPath.push(bmsvc.getItemTitle(afolderId)); 
	}

	// If it is an attribute folder stop the loop : the last afolderID is an attribute folder
	boucle = (isAttributeFolder(afolderId)) ? false : true;

	while (boucle){
	    afolderId = bmsvc.getFolderIdForItem(afolderId);
	    folderPath.push(bmsvc.getItemTitle(afolderId)); 
	    // If it is an attribute folder stop the loop : the last afolderID is an attribute folder
	    boucle = (isAttributeFolder(afolderId)) ? false : true;
	}
	folderPath.reverse();

	//alert("folderPath : "+folderPath.join(" > "));
	return folderPath.join(" > ");
    },

    // Function used to test if the bookmark is in Tag folder
    isItaTag: function (aItemID){
	var IdtoTest = aItemID;
	// If it is an attribute folder stop the loop : the last afolderID is an attribute folder
	var boucle = (isAttributeFolder(aItemID)) ? false : true;
	var result = false;

	while (boucle){
	    IdtoTest = bmsvc.getFolderIdForItem(IdtoTest);
	    // If it is an attribute folder stop the loop : the last afolderID is an attribute folder
	    boucle = (isAttributeFolder(IdtoTest)) ? false : true;
	}
	if (IdtoTest == bmsvc.tagsFolder) {
	    result=true;
	}
	if(Bookmarkdd.DEBUG) { Bookmarkdd.myDump("itemId :"+aItemID+" isItaTag = "+result+"\nPath='"+Bookmarkdd.findPath(aItemID)+"'"); }
	return result;
    },


    //Function used to log message within JS Console
    myDump: function (aMessage) 
    {
	var consoleService = Components.classes["@mozilla.org/consoleservice;1"]
				 .getService(Components.interfaces.nsIConsoleService);
	consoleService.logStringMessage("BookmarkDD's message: " + aMessage);
    }

};

// Test to know if argument is an 'attribute folder' of Bookmarks service
// Attributes folder are : bookmarksMenuFolder, tagsFolder, unfiledBookmarksFolder, toolbarFolder and the parent placesRoot
// return true or false
function isAttributeFolder(aFolderId){
    if (aFolderId == bmsvc.unfiledBookmarksFolder || 
	aFolderId == bmsvc.bookmarksMenuFolder || 
	aFolderId == bmsvc.toolbarFolder ||
	aFolderId == bmsvc.tagsFolder || 
	aFolderId == bmsvc.placesRoot ) {
	return true;
    } else {
	return false;
    }
}

//BookmarkDD uses its own confirm window to replace default icon without overlay icon for all Browser's confirm
// Params :
// aText = message to show in dialog box
// aModal = the dialog Box is modal or not
// By default it is not modal
function BddConfirm(aText,aModal){

    //Bookmarkdd.myDump("aModal="+aModal+"\naText = "+aText);
    if (aModal) {
	window.openDialog("chrome://duplicates/content/ConfirmOverlay.xul", "BookmarkDD-Confirm", "chrome,modal,alwaysRaised,resizable,centerscreen", aText);
    } else {
	window.openDialog("chrome://duplicates/content/ConfirmOverlay.xul", "BookmarkDD-Confirm", "chrome,alwaysRaised,resizable,centerscreen", aText);
    }

    return Bookmarkdd.userResponse;
}

// Overwrite default Firefox function which do 'sortFolderByName'
// It is needed to avoid detection of duplicates during the process of this function
if (typeof(PlacesController) != "undefined") {
    PlacesController.prototype.sortFolderByName = function PC_BddSortByNames() {
	Application.storage.set("Bdd_detectiondisabled", true);
	try {
	    //Original code taken from browser/places/controller.js
	    var itemId = PlacesUtils.getConcreteItemId(this._view.selectedNode);
	    var txn = PlacesUIUtils.ptm.sortFolderByName(itemId);
	    PlacesUIUtils.ptm.doTransaction(txn);
	} catch(err) { }
	Application.storage.set("Bdd_detectiondisabled", false);
    }
}

